from abc import ABC

import numpy as np
from braket.default_simulator.operation_helpers import from_braket_instruction
from braket.default_simulator.result_types import TargetedResultType
from braket.default_simulator.simulator import BaseLocalSimulator
from braket.device_schema import DeviceActionType
from braket.ir.jaqcd import DensityMatrix
from braket.ir.jaqcd import Program as JaqcdProgram
from braket.ir.jaqcd import StateVector
from braket.ir.openqasm import Program as OpenQASMProgram
from braket.task_result import GateModelTaskResult

from braket.simulator_v2.julia_import import jl


class BaseLocalSimulatorV2(BaseLocalSimulator, ABC):

    def __init__(self, device):
        self._device = device

    def initialize_simulation(self, **kwargs):
        return

    def run_jaqcd(
        self,
        circuit_ir: JaqcdProgram,
        qubit_count: int,
        shots: int = 0,
        batch_size: int = 1,  # unused
    ) -> GateModelTaskResult:
        """Executes the circuit specified by the supplied `circuit_ir` on the simulator.

        Args:
            circuit_ir (Program): ir representation of a braket circuit specifying the
                instructions to execute.
            qubit_count (int): The number of qubits to simulate.
            shots (int): The number of times to run the circuit.

        Returns:
            GateModelTaskResult: object that represents the result

        Raises:
            ValueError: If result types are not specified in the IR or sample is specified
                as a result type when shots=0. Or, if StateVector and Amplitude result types
                are requested when shots>0.
        """
        self._validate_ir_results_compatibility(
            circuit_ir.results,
            device_action_type=DeviceActionType.JAQCD,
        )
        self._validate_ir_instructions_compatibility(
            circuit_ir,
            device_action_type=DeviceActionType.JAQCD,
        )
        BaseLocalSimulator._validate_shots_and_ir_results(
            shots, circuit_ir.results, qubit_count
        )

        operations = [
            from_braket_instruction(instruction)
            for instruction in circuit_ir.instructions
        ]

        if shots and circuit_ir.basis_rotation_instructions:
            for instruction in circuit_ir.basis_rotation_instructions:
                operations.append(from_braket_instruction(instruction))

        BaseLocalSimulator._validate_operation_qubits(operations)

        if not shots and circuit_ir.results:
            result_types = BaseLocalSimulator._translate_result_types(
                circuit_ir.results
            )
            BaseLocalSimulator._validate_result_types_qubits_exist(
                [
                    result_type
                    for result_type in result_types
                    if isinstance(result_type, TargetedResultType)
                ],
                qubit_count,
            )
        r = jl.simulate(self._device, [circuit_ir], qubit_count, shots)
        r.additionalMetadata.action = circuit_ir
        if not shots:
            r = _analytic_result_value_to_ndarray(r)
        return r

    def run_openqasm(
        self,
        openqasm_ir: OpenQASMProgram,
        shots: int = 0,
        batch_size: int = 1,  # unused
    ) -> GateModelTaskResult:
        """Executes the circuit specified by the supplied `openqasm_ir` on the simulator.

        Args:
            openqasm_ir (Program): ir representation of a braket circuit specifying the
                instructions to execute.
            shots (int): The number of times to run the circuit.

        Returns:
            GateModelTaskResult: object that represents the result

        Raises:
            ValueError: If result types are not specified in the IR or sample is specified
                as a result type when shots=0. Or, if StateVector and Amplitude result types
                are requested when shots>0.
        """

        # use Python parser
        circuit = self.parse_program(openqasm_ir).circuit
        qubit_count = circuit.num_qubits
        measured_qubits = circuit.measured_qubits

        self._validate_ir_results_compatibility(
            circuit.results,
            device_action_type=DeviceActionType.OPENQASM,
        )
        self._validate_ir_instructions_compatibility(
            circuit,
            device_action_type=DeviceActionType.OPENQASM,
        )
        self._validate_input_provided(circuit)
        BaseLocalSimulator._validate_shots_and_ir_results(
            shots, circuit.results, qubit_count
        )

        operations = circuit.instructions
        BaseLocalSimulator._validate_operation_qubits(operations)

        results = circuit.results

        if not shots:
            result_types = BaseLocalSimulator._translate_result_types(results)
            BaseLocalSimulator._validate_result_types_qubits_exist(
                [
                    result_type
                    for result_type in result_types
                    if isinstance(result_type, TargetedResultType)
                ],
                qubit_count,
            )
        else:
            for bri in circuit.basis_rotation_instructions:
                circuit.add_instruction(bri)

        r = jl.simulate(
            self._device, [circuit], qubit_count, shots, measured_qubits=measured_qubits
        )
        r.additionalMetadata.action = openqasm_ir
        # attach the result types
        if shots:
            r.resultTypes = results
        else:
            r = _analytic_result_value_to_ndarray(r)
        return r


def _analytic_result_value_to_ndarray(
    task_result: GateModelTaskResult,
) -> GateModelTaskResult:
    """Convert any StateVector or DensityMatrix result values from raw Python lists to the expected
    np.ndarray. This must be done because the wrapper Julia simulator results Python lists to comply
    with the pydantic specification for ResultTypeValues.
    """
    for result_ind, result_type in enumerate(task_result.resultTypes):
        if isinstance(result_type.type, StateVector) or isinstance(
            result_type.type, DensityMatrix
        ):
            task_result.resultTypes[result_ind].value = np.asarray(
                task_result.resultTypes[result_ind].value
            )
    return task_result



import os
import sys
import warnings

# Check if JuliaCall is already loaded, and if so, warn the user
# about the relevant environment variables. If not loaded,
# set up sensible defaults.
if "juliacall" in sys.modules:
    warnings.warn(
        "`juliacall` module has already been imported. "
        + "Make sure that you have set the environment variable "
        + "`PYTHON_JULIACALL_HANDLE_SIGNALS=yes` to avoid segfaults. "
    )
else:
    # Required to avoid segfaults (https://juliapy.github.io/PythonCall.jl/dev/faq/)
    if os.environ.get("PYTHON_JULIACALL_HANDLE_SIGNALS", "yes") != "yes":
        warnings.warn(
            "`PYTHON_JULIACALL_HANDLE_SIGNALS` environment variable "
            + "is set to something other than 'yes' or ''. "
            + +"You will experience segfaults if running with Julia multithreading."
        )

    if os.environ.get("PYTHON_JULIACALL_THREADS", "auto") != "auto":
        warnings.warn(
            "`PYTHON_JULIACALL_THREADS` environment variable is set to "
            + "something other than `auto`, so `amazon-braket-simulator-v2` "
            + "was not able to set it."
        )

    for k, default in (
        ("PYTHON_JULIACALL_HANDLE_SIGNALS", "yes"),
        ("PYTHON_JULIACALL_THREADS", "auto"),
        ("PYTHON_JULIACALL_OPTLEVEL", "3"),
        # let the user's Conda/Pip handle installing things
        ("JULIA_CONDAPKG_BACKEND", "Null"),
    ):
        os.environ[k] = os.environ.get(k, default)

import braket.default_simulator.gate_operations as gates
import braket.default_simulator.noise_operations as noises
from braket.default_simulator.openqasm.circuit import Circuit
import juliacall

jl = juliacall.Base.Module()

jl.seval("using PythonCall, BraketSimulator")
jl.seval("using PythonCall: Py, pyconvert")
jlBraketSimulator = jl.BraketSimulator

# at init run many commonly used gates and noises to force compilation
# for this Python session
# must be done at init because each new Python session has a separate handle and
# seems to invalidate old methods

circuit = Circuit()
for instruction in (
    gates.Hadamard(targets=[0]),
    gates.RotX(targets=[0], angle=0.1),
    gates.RotY(targets=[1], angle=0.2),
    gates.RotZ(targets=[0], angle=0.1),
    gates.PhaseShift(targets=[0], angle=0.1),
    gates.PRx(targets=[0], angle_1=0.1, angle_2=0.2),
    gates.GPi(targets=[0], angle=0.1),
    gates.GPi2(targets=[0], angle=0.1),
    gates.PauliX(targets=[0]),
    gates.PauliY(targets=[1]),
    gates.PauliZ(targets=[0]),
    gates.V(targets=[0]),
    gates.Vi(targets=[1]),
    gates.S(targets=[0]),
    gates.Si(targets=[1]),
    gates.T(targets=[0]),
    gates.Ti(targets=[1]),
    gates.Identity(targets=[1]),
    gates.Swap(targets=[0, 1]),
    gates.ISwap(targets=[0, 1]),
    gates.PSwap(targets=[0, 1], angle=0.1),
    gates.ECR(targets=[0, 1]),
    gates.CX(targets=[0, 1]),
    gates.CY(targets=[0, 1]),
    gates.CZ(targets=[0, 1]),
    gates.CV(targets=[0, 1]),
    gates.MS(targets=[0, 1], angle_1=0.1, angle_2=0.2, angle_3=0.3),
    gates.XX(targets=[0, 1], angle=0.1),
    gates.XY(targets=[0, 1], angle=0.1),
    gates.YY(targets=[0, 1], angle=0.1),
    gates.ZZ(targets=[0, 1], angle=0.1),
    gates.CPhaseShift(targets=[0, 1], angle=0.1),
    gates.CPhaseShift00(targets=[0, 1], angle=0.1),
    gates.CPhaseShift10(targets=[0, 1], angle=0.1),
    gates.CPhaseShift01(targets=[0, 1], angle=0.1),
    gates.CCNot(targets=[0, 1, 2]),
    gates.CSwap(targets=[0, 1, 2]),
):
    circuit.add_instruction(instruction)

jl.simulate(
    jlBraketSimulator.StateVectorSimulator(0, 0), [circuit], 3, 100, measured_qubits=[]
)

circuit = Circuit()
for instruction in (
    gates.Hadamard(targets=[0]),
    gates.CX(targets=[0, 1]),
    noises.BitFlip(targets=[0], probability=0.1),
    noises.PhaseFlip(targets=[0], probability=0.1),
    noises.PauliChannel(targets=[0], probX=0.1, probY=0.2, probZ=0.3),
    noises.Depolarizing(targets=[0], probability=0.1),
    noises.TwoQubitDepolarizing(targets=[0, 1], probability=0.1),
    noises.TwoQubitDephasing(targets=[0, 1], probability=0.1),
    noises.AmplitudeDamping(targets=[0], gamma=0.1),
    noises.PhaseDamping(targets=[0], gamma=0.1),
    noises.GeneralizedAmplitudeDamping(targets=[0], gamma=0.1, probability=0.2),
):
    circuit.add_instruction(instruction)

jl.simulate(
    jlBraketSimulator.DensityMatrixSimulator(0, 0),
    [circuit],
    3,
    100,
    measured_qubits=[],
)
